﻿package com.selenetan.core 
{
    import flash.display.Sprite;
    import flash.geom.Point;
	/**
     * ...
     * @author Selene
     */
    public class STObject
    {
        /**
         * The (Flash) sprite displayed for this object.
         */
        public var sprite:Sprite;
        
        /**
         * The sprite used for first-pass hit detection. Further detection uses circles. May be the same as sprite.
         */
        public var hitSprite:Sprite;
        
        /**
         * The hit radius of the sprite. Defaults to half the width of the hitSprite.
         */
        public var hitRadius:Number;
        
        /**
         * True if this object is available to be re-used.
         */
        public var exists:Boolean;
        
        /**
         * True if this object is dead.
         */
        public var dead:Boolean;
        
        /**
         * True if this object is being rendered.
         */
        public var visible:Boolean;
        
        /**
         * Object's remaining health. Really, this is just a number, though.
         */
        public var health:Number;
        
        // ---------------------
        // Physics-y properties:
        
        /**
         * If this is true, the object does not move when it collides.
         */
        public var fixed:Boolean;
        
        public var x:Number;
        public var y:Number;
        
        public var velocity:Point;
        public var maxVelocity:Point;
        public var acceleration:Point;
        public var maxAcceleration:Point;
        
        /**
         * Drag applied to the object. This number is multiplied by velocity every frame.
         */
        public var linearDrag:Number;
        
        /**
         * Angle in degrees, using math conventions. Angle 0 points to the right. Positive angles are counter-clockwise from 0.
         * When updating the sprite, rotation is calculated so that angle 0 points up and positive angles are clockwise from 0.
         * (Graphics/animation conventions.)
         */
        public var angle:Number;
        public var angularVelocity:Number;
        
        /**
         * Acceleration in the direction of current facing. It is added to acceleration values.
         */
        public var thrust:Number;
        public var maxThrust:Number;
        
        /**
         * A measure of how sticky/bouncy this object is. Used to calculate the coefficient of restitution in collisions.
         * If both objects in a collision have elasticitty 1, the collision will be elastic.
         */
        public var elasticity:Number;
        public var mass:Number;
        
        protected var _ticker:STTicker
        
        /**
         * An object with some basic physics and an attached Flash Sprite
         * @param	$x              Starting x-coordinate
         * @param	$y              Starting y-coordinate
         * @param	$sprite         The display sprite for this object.
         * @param	$hitSprite=null A sprite to use for collision detection. If null, it will default to sprite.
         */
        public function STObject($sprite:Sprite, $hitSprite:Sprite=null, $x:Number=0, $y:Number=0, $ticker:STTicker=null) 
        {
            x = $x;
            y = $y;
            sprite = $sprite;
            if ($hitSprite == null) {
                hitSprite = sprite;
            }
            else {
                hitSprite = $hitSprite;
            }
            
            hitRadius = hitSprite.width * 0.5;
            
            if ($ticker == null) {
                _ticker = STGlobal.ticker;
            }
            
            exists = true;
            dead = false;
            visible = true;
            
            fixed = false;
            
            velocity = new Point();
            maxVelocity = new Point(100000, 100000);
            acceleration = new Point();
            maxAcceleration = new Point(100000, 100000);
            
            linearDrag = 1;
            
            angle = 0;
            angularVelocity = 0;
            
            thrust = 0;
            maxThrust = 100000;
            
            elasticity = 1;
            mass = 1;
        }
        
        
        /**
         * Called every frame to update logic and physics
         */
        public function update():void
        {
            updateMotion();
        }
        
        /**
         * Update position and angle based on physics
         */
        protected function updateMotion():void
        {
            var elapsed:Number = _ticker.elapsed;
            
            angle += angularVelocity * elapsed;
            
            if (thrust != 0) {
                if (thrust > maxThrust) {
                    thrust = maxThrust;
                }
                else if (thrust < -maxThrust) {
                    thrust = -maxThrust;
                }
                
                // Y acceleration is negative because screen coordinates have up/down reversed compared to Cartesian coordinates
                acceleration.x += thrust * Math.cos(angle * Math.PI / 180.0) * elapsed;
                acceleration.y -= thrust * Math.sin(angle * Math.PI / 180.0) * elapsed;
            }
            
            acceleration.x = STUtil.clamp(acceleration.x, maxAcceleration.x);
            acceleration.y = STUtil.clamp(acceleration.y, maxAcceleration.y);
            
            velocity.x += acceleration.x * elapsed * 0.5;
            velocity.y += acceleration.y * elapsed * 0.5;
            
            velocity.x *= linearDrag;
            velocity.y *= linearDrag;
            
            velocity.x = STUtil.clamp(velocity.x, maxVelocity.x);
            velocity.y = STUtil.clamp(velocity.y, maxVelocity.y);
                        
            x += velocity.x * elapsed;
            y += velocity.y * elapsed;
            
            velocity.x += acceleration.x * elapsed * 0.5;
            velocity.y += acceleration.y * elapsed * 0.5;
            
            if (hitSprite != sprite) {
                hitSprite.x = x;
                hitSprite.y = y;
                hitSprite.rotation = 90 - angle;
            }
        }
        
        
        /**
         * Called every frame to update visuals. If visible is false, this function bails early.
         */
        public function render():void
        {
            sprite.visible = visible;
            if (!visible) {
                return;
            }
            
            sprite.x = x;
            sprite.y = y;
            
            sprite.rotation = 90 - angle;
            
        }
        
        /**
         * Apply damage to the object's health.  Negative damage will heal the object. If health drops below 0, call kill().
         * @param	damage  Amount of damage to apply.
         */
        public function hurt(damage:Number):void
        {
            health -= damage;
            
            if (health < 0) {
                kill();
            }
        }
        
        /**
         * Marks the sprite as dead, non-existing, and invisible.
         */
        public function kill():void
        {
            dead = true;
            exists = false;
            visible = false;
        }
        
        
        /**
         * A basic reset of the object's parameters. Marks it as alive, existing, and visible.
         * @param	$x      New x-coordinate
         * @param	$y      New y-coordinate
         * @param	$args   Pass extra arguments to overridden versions of this function.
         */
        public function reset($x:Number = 0, $y:Number = 0, $args:Object = null):void
        {
            x = $x;
            y = $y;
            
            dead = false;
            exists = true;
            visible = true;
        }
        
    }

}